
import psycopg2
from psycopg2.extras import DictCursor as _CURSOR_FACTORY

# Importing db connection configuration
from conf import\
    DB_HOST,DB_NAME,DB_USER,DB_PASSWD

from speaker import *
from eccez import *
#
##
### EXPORTING INTERFACE
##
#

_DBDA_ATTRS=['commit',\
             'mogrify',\
             'ov_query',\
             'rollback',\
             'oc_query',\
             'or_query',\
             'mr_query',\
             'beginTrans',\
             'lastrowid']

def dbda_iface_init(c_to,c_from):
    for a in _DBDA_ATTRS:
        na= getattr(c_from, a, lambda: None)
        setattr(c_to, a, na)

def _connectStr():
    return DB_HOST and\
        "dbname=%s user=%s host=%s password=%s"\
            % (DB_NAME,DB_USER,DB_HOST,DB_PASSWD) or\
        "dbname=%s user=%s password=%s"\
            % (DB_NAME,DB_USER,DB_PASSWD) 
#
##
### The DbDa ClaSs
##
#

_DB_CONN=None
def dbda_init(db_conn=None):
    global _DB_CONN
    _DB_CONN=db_conn


class DbDa:
    def __init__(self, logger= None):

        self.dbda=self.db_cursor=None
        self.transaction=0
        if logger is not None:
            speaker_iface_init(self,logger)
        
    def _connect(self):
        if _DB_CONN is None:
            self.log_debug("Connecting to db ...",DBG_INSANE)
            self.dbda=psycopg2.connect(_connectStr())
            self.db_cursor=self.dbda.cursor(\
                cursor_factory=_CURSOR_FACTORY)
        else:
            self.log_debug("Using pre-configured DBDA ...",DBG_INSANE)
            self.dbda=_DB_CONN
            self.db_cursor=self.dbda.getcursor()

        self._replaceQryFunctions()

    def _replaceQryFunctions(self):
        self.log_debug("Switching functions ...",DBG_INSANE)
        self._get_query=self.__get_query
        self._quiet_query=self.__quiet_query
        self.mogrify=self.__mogrify
        self.beginTrans= self.__beginTrans

#
##
### INTERNALS
##
#
    def _get_query(self,s):
        self._connect()
        return self.__get_query(s)

    def _quiet_query(self,s):
        self._connect()
        return self.__quiet_query(s)

    def __get_query(self,s):
        self.log_debug("QSQL- %s",s,DBG_NORMAL)
        try:
            self.db_cursor.execute(s)
            res=self.db_cursor.fetchall()
        except Exception,inst:
            self.dbda.rollback()
            s="Unable to execute query: %s" % str(inst)
            self.log_raise(s,QueryError(s))
        return res

    def __quiet_query(self,s):
        self.log_debug("QSQL- %s",s,DBG_NORMAL)
        try:
            self.db_cursor.execute(s)
            if not self.transaction:
                self.dbda.commit()
            else:
                self.log_debug("Query not committed.",DBG_INSANE)
        except Exception,inst:
            self.dbda.rollback()
            s="Unable to execute query: %s" % str(inst)
            self.log_raise(s,QueryError(s))

    def _setTrans(self):
        self.transaction+=1
        self.log_debug("Incremented transaction: level=%s.",\
                YELLOW(self.transaction),DBG_INSANE)

    def _unsetTrans(self):
        self.transaction-=1
        self.log_debug("Decremented transaction: level=%s.",\
                YELLOW(self.transaction),DBG_INSANE)

    def _resetTrans(self):
        self.log_debug("Resetting transaction at level 0.",DBG_INSANE)
        self.transaction=0
#
##
### PUB
##
#
    def __beginTrans(self):
        if self.transaction:
            self.log_debug("Nested transaction, level=%s.",\
                    YELLOW(self.transaction),DBG_SOFT)
        self.transaction+=1
        self.log_debug("Incremented transaction: level=%s.",\
                YELLOW(self.transaction),DBG_INSANE)
        return
        self._setTrans()

    def beginTrans(self):
        self._connect()
        self.__beginTrans()

    def commit(self):
        if not self.transaction:
            self.log_int_raise("What do you want to commit? "\
                    "There is no active transaction.")
        self._unsetTrans()
        if self.transaction:
            self.log_debug("Transaction is not at level 0, not committing.",\
                    DBG_SOFT)
        else:
            self.dbda.commit()
	    self.log_debug("Queries committed.",DBG_SOFT)

    def rollback(self):
        if not self.transaction:
            self.log_debug("There is no active transaction. Maybe "\
                    "rollback was called nested?",DBG_SOFT)
        else:
            self._resetTrans()
            self.dbda.rollback()
            self.log_debug("Rollbacked changes.",DBG_SOFT)

    def __mogrify(self,s,d):
        """ Mix a string and a dict: see cursor.mogrify."""
        s=self.db_cursor.mogrify(s,d)
        return s

    def mogrify(self,s,d):
        self._connect()
        return self.__mogrify(s,d)

    def lastrowid(self):
        return self.db_cursor.lastrowid

    def oc_query(self,s):
        """ One Commit Query: no returns."""
        self._quiet_query(s)

    def ov_query(self,s):
        """ One Value Query: returns one only value."""
        res=self._get_query(s)
        try:
            return res[0][0]
        except:
            self.log_int_raise(
                "No value for the executed OV-query.")

    def or_query(self,s):
        """ One Record Query: returns a dict."""
        res=self._get_query(s)
        try:
            res=res[0]
        except:
            self.log_int_raise(
               "No values for the query.")
        d={}
        for k in res.keys():
            d[k]=res[k]
        return d

    def mr_query(self,s):
        """ Multiple Records Query: return a dict list."""
        res=self._get_query(s)
        dict_list=[]
        for r in res:
            d={}
            for key in r.keys():
                d[key]=r[key]
            dict_list.append(d)
        return dict_list
